import importlib.resources as pkg_resources
import os
import re
from typing import Dict, List, Optional, Tuple, Union

import awswrangler as wr
import boto3
from boto3.session import Session
from dataengineeringutils3.s3 import read_json_from_s3
from jinja2 import Template

from mojap_metadata.converters.aws_iceberg_converter import specs
from mojap_metadata.converters.aws_iceberg_converter.exceptions import (
    GlueIcebergTableExists,
    MalformedIcebergPartition,
    NonIcebergGlueTable,
    UnsupportedIcebergSchemaEvolution,
)
from mojap_metadata.converters.aws_iceberg_converter.iceberg_metadata import (
    IcebergMetadata,
)
from mojap_metadata.converters.glue_converter import GlueConverter, GlueTable
from mojap_metadata.metadata.metadata import Metadata

_supported_changes = [
    (
        "integer",
        "bigint",
    ),
    (
        "int",
        "bigint",
    ),
    (
        "float",
        "double",
    ),
]


class AthenaIcebergSqlConverter(GlueConverter):
    @staticmethod
    def _combine_partition_name_and_transform(name: str, transform: str) -> str:
        transform_arg_search = re.search("\\[[0-9]*\\]", transform)

        if "bucket" in transform or "truncate" in transform:
            if transform_arg_search is not None:
                transform_arg = re.sub("\\[|\\]", "", transform_arg_search.group(0))
                transform_fn = re.sub("\\[.*\\]", "", transform)

            else:
                raise MalformedIcebergPartition(
                    f"{transform} should have a length / width variable specified"
                )

        if "bucket" in transform:
            full_partition = f"{transform_fn}({transform_arg}, {name})"
        elif "truncate" in transform:
            full_partition = f"{transform_fn}({name}, {transform_arg})"
        elif transform == "identity":
            full_partition = name
        else:
            full_partition = f"{transform}({name})"

        return full_partition

    def convert_columns(self, metadata: IcebergMetadata) -> Tuple[List, List]:
        cols = [
            (
                c["name"],
                self.convert_col_type(c["type"]),
                c["description"] if "description" in c else "",
            )
            for c in metadata.columns
        ]

        if metadata.partition_transforms is not None:
            partitions = [
                self._combine_partition_name_and_transform(
                    name,
                    transform,
                )
                for name, transform in zip(
                    metadata.partitions,
                    metadata.partition_transforms,
                )
            ]
        else:
            partitions = []

        return cols, partitions

    def generate_from_meta(
        self,
        metadata: Union[IcebergMetadata, str, dict],
        database_name: str = None,
        table_location: str = None,
        create_not_alter: bool = True,
        existing_metadata: Optional[Union[IcebergMetadata, str, dict, None]] = None,
        **kwargs,
    ) -> List[str]:
        method_id = "create" if create_not_alter else "alter"
        method = getattr(self, f"generate_{method_id}_from_meta")

        metadata_to_use = IcebergMetadata.from_infer(metadata)
        table_location_to_use = (
            table_location if table_location is not None else metadata.table_location
        )
        database_name_to_use = (
            database_name if database_name is not None else metadata.database_name
        )

        if database_name_to_use is None:
            raise ValueError("Database name must be set in metadata or via method")

        method_kwargs = {
            "metadata": metadata_to_use,
            "database_name": database_name_to_use,
            "table_location": table_location_to_use,
            **kwargs,
        }

        if not create_not_alter:
            if existing_metadata is None:
                raise ValueError("existing_metadata must be specified")

            method_kwargs["existing_metadata"] = existing_metadata

        queries = method(**method_kwargs)
        if isinstance(queries, str):
            queries = [queries]

        return queries

    def generate_create_from_meta(
        self,
        metadata: Union[IcebergMetadata, str, dict],
        database_name: str = None,
        table_location: str = None,
        **kwargs,
    ) -> str:
        metadata_to_use = IcebergMetadata.from_infer(metadata)
        table_location_to_use = (
            table_location if table_location is not None else metadata.table_location
        )
        database_name_to_use = (
            database_name if database_name is not None else metadata.database_name
        )

        if table_location_to_use is None:
            raise ValueError("Table location must be set in metadata or via method")

        columns, partitions = self.convert_columns(metadata_to_use)

        raw_sql = pkg_resources.open_text(specs, "create.sql").read()

        sql_template = Template(raw_sql)

        sql = sql_template.render(
            table_name=metadata_to_use.name,
            database_name=database_name_to_use,
            table_location=table_location_to_use,
            columns=columns,
            partitions=partitions,
            **kwargs,
        )

        return sql

    def generate_alter_from_meta(
        self,
        metadata: Union[IcebergMetadata, str, dict],
        existing_metadata: Union[IcebergMetadata, str, dict],
        database_name: str = None,
        **kwargs,
    ) -> List[str]:
        meta = IcebergMetadata.from_infer(metadata)
        existing_meta = IcebergMetadata.from_infer(existing_metadata)

        database_name_to_use = (
            database_name if database_name is not None else metadata.database_name
        )

        add_columns, changed_columns = self._find_new_or_updated_columns(
            metadata=meta,
            existing_metadata=existing_meta,
        )

        removed_columns = self._find_removed_columns(
            metadata=meta,
            existing_metadata=existing_meta,
        )

        queries = []
        if add_columns:
            add_query = self._generate_alter_add_query(
                columns=add_columns,
                table_name=meta.name,
                database_name=database_name_to_use,
                **kwargs,
            )
            queries.append(add_query)

        queries += self._generate_alter_change_queries(
            changed_columns,
            table_name=meta.name,
            database_name=database_name_to_use,
        )

        queries += self._generate_alter_drop_queries(
            removed_columns,
            table_name=meta.name,
            database_name=database_name_to_use,
        )

        return queries

    @staticmethod
    def _generate_alter_add_query(
        columns: List[Tuple[str, str]], table_name: str, database_name: str, **kwargs
    ) -> str:
        raw_sql = pkg_resources.open_text(specs, "alter_add_columns.sql").read()

        sql_template = Template(raw_sql)

        sql = sql_template.render(
            table_name=table_name,
            database_name=database_name,
            columns=columns,
            **kwargs,
        )

        return sql

    @staticmethod
    def _generate_alter_change_queries(
        changed_columns: List[Tuple[str, str]],
        table_name: str,
        database_name: str,
        **kwargs,
    ) -> List[str]:
        alter_change_queries = []

        raw_sql = pkg_resources.open_text(specs, "alter_change_column.sql").read()

        sql_template = Template(raw_sql)

        for column_name, column_type in changed_columns:
            sql = sql_template.render(
                table_name=table_name,
                database_name=database_name,
                column_old_name=column_name,
                column_new_name=column_name,
                column_type=column_type,
                **kwargs,
            )
            alter_change_queries.append(sql)

        return alter_change_queries

    @staticmethod
    def _generate_alter_drop_queries(
        removed_columns: List[str], table_name: str, database_name: str, **kwargs
    ) -> List[str]:
        alter_drop_queries = []

        raw_sql = pkg_resources.open_text(specs, "alter_drop_column.sql").read()

        sql_template = Template(raw_sql)

        for column_name in removed_columns:
            sql = sql_template.render(
                table_name=table_name,
                database_name=database_name,
                column_name=column_name,
                **kwargs,
            )
            alter_drop_queries.append(sql)

        return alter_drop_queries

    def _find_new_or_updated_columns(
        self, metadata: Metadata, existing_metadata: Metadata
    ) -> List[Tuple[str, str]]:
        add_columns = []
        changed_columns = []

        for column in metadata.columns:
            column_name = column["name"]
            column_type = self.convert_col_type(column["type"])

            if column_name not in existing_metadata.column_names:
                add_columns.append((column_name, column_type))

            else:
                old_column_type = self.convert_col_type(
                    existing_metadata.get_column(column_name)["type"]
                )

                if column_type != old_column_type:
                    supported = (old_column_type, column_type) in _supported_changes

                    if not supported:
                        raise UnsupportedIcebergSchemaEvolution(
                            f"Can't implement change from {old_column_type}"
                            + f" to {column_type}"
                        )

                    changed_columns.append((column_name, column_type))

        return add_columns, changed_columns

    @staticmethod
    def _find_removed_columns(
        metadata: Metadata, existing_metadata: Metadata
    ) -> List[str]:
        return [
            column_name
            for column_name in existing_metadata.column_names
            if column_name not in metadata.column_names
        ]


class AwsIcebergTable(GlueTable):
    def __init__(self):
        super().__init__(glue_converter_options=None)
        self.sql_converter = AthenaIcebergSqlConverter()

    @staticmethod
    def _pre_generation_setup(
        database_name: str,
        table_name: str,
        delete_table_if_exists: bool,
        table_location: Union[str, None] = None,
        boto3_session: Union[Session, None] = None,
    ) -> bool:
        # create database if it doesn't exist
        existing_databases = wr.catalog.databases(None).Database.to_list()
        if database_name not in existing_databases:
            wr.catalog.create_database(name=database_name, boto3_session=boto3_session)

        if delete_table_if_exists:
            if table_location is None:
                raise ValueError("Table location must be set to re-create the table")

            _ = wr.catalog.delete_table_if_exists(
                database=database_name, table=table_name, boto3_session=boto3_session
            )
            wr.s3.delete_objects(table_location, boto3_session=boto3_session)
            table_exists = False

        else:
            table_exists = wr.catalog.does_table_exist(
                database=database_name, table=table_name, boto3_session=boto3_session
            )

        return table_exists

    @staticmethod
    def _execute_queries(
        queries: List[str], boto3_session: Union[Session, None] = None
    ):
        for query in queries:
            try:
                query_id = wr.athena.start_query_execution(
                    query, boto3_session=boto3_session
                )
                response = wr.athena.wait_query(
                    query_execution_id=query_id, boto3_session=boto3_session
                )
            except Exception:
                raise wr.exceptions.QueryFailed(f"Failed to execute:\n{query}")

            if response["Status"]["State"] == "FAILED":
                raise wr.exceptions.QueryFailed(f"Failed to execute:\n{query}")

            if response["Status"]["State"] == "CANCELLED":
                raise wr.exceptions.QueryCancelled(f"Query cancelled:\n{query}")

    def generate_from_meta(
        self,
        metadata: Union[Metadata, str, dict],
        database_name: str = None,
        table_location: str = None,
        delete_table_if_exists: Optional[bool] = False,
        alter_table_if_exists: Optional[bool] = False,
        boto3_session: Union[Session, None] = None,
    ):
        """
        Creates an Iceberg table in AWS from metadata
        arguments:
            - metadata: Metadata object, string path, or dictionary metadata.
            - database_name (optional): name of the glue database the table is to be
            created in. can also be a property of the metadata.
            - table_location (optional): the s3 location of the table. can also be a
            property of the metadata.
            - delete_table_if_exists (optional): whether to delete the iceberg table and
            underlying data if it exists. Defaults to False.
            - alter_table_if_exists (optional): whether to alter the iceberg table
            schema if it already exists. Defaults to False.
        """

        # set database_name to metadata.database_name if none
        database_name = database_name if database_name else metadata.database_name
        # do the same with table_location
        table_location = table_location if table_location else metadata.table_location

        if database_name is None:
            raise ValueError("Database name must be set in metadata or via method")

        metadata = Metadata.from_infer(metadata)

        table_exists = self._pre_generation_setup(
            database_name=database_name,
            table_name=metadata.name,
            table_location=table_location,
            delete_table_if_exists=delete_table_if_exists,
            boto3_session=boto3_session,
        )

        if not table_exists or alter_table_if_exists:
            queries = self.sql_converter.generate_from_meta(
                metadata=metadata,
                database_name=database_name,
                table_location=table_location,
                create_not_alter=not table_exists,
                existing_metadata=self.generate_to_meta(
                    database=database_name,
                    table=metadata.name,
                )
                if alter_table_if_exists
                else None,
            )

            _ = self._execute_queries(queries, boto3_session=boto3_session)

        else:
            raise GlueIcebergTableExists(
                f"{metadata.name} already exists and delete_table_if_exists"
                + " and alter_table_if_exists both set to False"
            )

    def generate_to_meta(self, database: str, table: str) -> IcebergMetadata:
        glue_client = boto3.client(
            "glue",
            region_name=os.getenv(
                "AWS_REGION", os.getenv("AWS_DEFAULT_REGION", "eu-west-1")
            ),
        )

        resp = glue_client.get_table(
            DatabaseName=database,
            Name=table,
        )

        table_details = resp["Table"]
        table_storage_details = table_details["StorageDescriptor"]

        table_parameters = table_details.get("Parameters", {})
        table_type = table_parameters.get("table_type", "").lower()

        if table_type != "iceberg":
            raise NonIcebergGlueTable(
                "Please specify an iceberg table registered with Glue"
            )

        table_metadata_location = table_parameters["metadata_location"]
        (
            table_partitions,
            table_partition_transforms,
        ) = self._get_partitions_from_iceberg_metadata(table_metadata_location)

        table_location = table_storage_details["Location"]
        table_iceberg_columns = table_storage_details["Columns"]

        table_current_columns = self.convert_columns(
            [
                {"Name": c["Name"], "Type": c["Type"]}
                for c in table_iceberg_columns
                if c["Parameters"]["iceberg.field.current"] == "true"
            ]
        )

        table_noncurrent_columns = self.convert_columns(
            [
                {"Name": c["Name"], "Type": c["Type"]}
                for c in table_iceberg_columns
                if c["Parameters"]["iceberg.field.current"] == "false"
            ]
        )

        meta_dict = {
            "columns": table_current_columns,
            "table_type": "iceberg",
            "file_format": "parquet",
            "table_location": table_location,
            "noncurrent_columns": table_noncurrent_columns,
            "partitions": table_partitions,
            "partition_transforms": table_partition_transforms,
            "name": table,
            "database_name": database,
        }

        meta = IcebergMetadata.from_dict(meta_dict)

        return meta

    @classmethod
    def _get_partitions_from_iceberg_metadata(
        cls,
        iceberg_metadata_filepath: str,
    ) -> List[str]:
        table_iceberg_metadata = read_json_from_s3(iceberg_metadata_filepath)

        table_partition_spec_id = table_iceberg_metadata.get("default-spec-id")
        table_partition_specs = table_iceberg_metadata.get("partition-specs", [])

        raw_partitions = cls._get_current_partition_spec(
            table_partition_specs, table_partition_spec_id
        )
        partitions = [p["name"] for p in raw_partitions]
        partiton_transforms = [p["transform"] for p in raw_partitions]

        return partitions, partiton_transforms

    @staticmethod
    def _get_current_partition_spec(
        partition_specs: List[Dict], current_partition_spec_id: int
    ) -> List[Dict]:
        partitions = []
        for partition_spec in partition_specs:
            partition_spec_id = partition_spec["spec-id"]
            if partition_spec_id == current_partition_spec_id:
                partitions += partition_spec["fields"]
        return partitions



from typing import Callable, List, Union
from copy import deepcopy
from mojap_metadata.metadata.metadata import (
    Metadata,
    _get_first_level,
    _parse_and_split,
    _unpack_complex_data_type,
)
from mojap_metadata.converters import (
    BaseConverter,
    _flatten_and_convert_complex_data_type,
)
import warnings
from etl_manager.meta import TableMeta

# Format generictype: (glue_type, is_fully_supported)
# If not fully supported we decide on best option
# if glue_type is Null then we have no way to safely
# convert it
_default_type_converter = {
    "null": (None, False),
    "bool": ("boolean", True),
    "bool_": ("boolean", True),
    "int8": ("int", False),
    "int16": ("int", False),
    "int32": ("int", True),
    "int64": ("long", True),
    "uint8": ("int", False),
    "uint16": ("int", False),
    "uint32": ("long", False),
    "uint64": (None, False),
    "decimal128": ("decimal", True),
    "float16": ("float", False),
    "float32": ("float", True),
    "float64": ("double", True),
    "time32": (None, False),
    "time32(s)": (None, False),
    "time32(ms)": (None, False),
    "time64(us)": (None, False),
    "time64(ns)": (None, False),
    "date32": ("date", True),
    "date64": ("date", True),
    "timestamp(s)": ("datetime", True),
    "timestamp(ms)": ("datetime", True),
    "timestamp(us)": ("datetime", True),
    "timestamp(ns)": ("datetime", True),
    "string": ("character", True),
    "large_string": ("character", True),
    "utf8": ("character", True),
    "large_utf8": ("character", True),
    "binary": ("binary", True),
    "large_binary": ("binary", True),
    "list_": ("array", True),
    "list": ("array", True),
    "large_list": ("array", True),
    "struct": ("struct", True),
}

_reverse_default_type_converter = {
    "character": ("string", True),
    "int": ("int32", True),
    "long": ("int64", True),
    "float": ("float32", True),
    "double": ("float64", True),
    "decimal": ("decimal128", True),
    "date": ("date32", True),
    "datetime": ("timestamp(s)", True),
    "binary": ("binary", True),
    "boolean": ("bool", True),
    "struct": ("struct", True),
    "array": ("list", True),
}


def _unpack_complex_etl_type(data_type: str) -> Union[str, dict]:
    """Recursive function that jumps into complex
    data types and returns complex types as a dict.
    Non complex types are returned as a str. Similar to
    mojap_metadata.metadata.metadata._unpack_complex_data_type
    but uses etl type names instead of agnostic names.

    Args:
        data_type (str): Name of etl-manager data type

    Returns:
        Union[str, dict]: unpacked representation of data type as dict
            for complex types. If datatype is not complex then original
            data type is returned (as str).
    """
    d = {}
    if data_type.startswith("struct<"):
        d["struct"] = {}
        next_data_type = _get_first_level(data_type)
        for data_param in _parse_and_split(next_data_type, ","):
            k, v = data_param.split(":", 1)
            k = k.strip()
            v = v.strip()
            if v.startswith("struct<") or v.startswith("array<"):
                d["struct"][k] = _unpack_complex_etl_type(v)
            else:
                d["struct"][k] = v
        return d
    elif data_type.startswith("array<"):
        d["array"] = {}
        next_data_type = _get_first_level(data_type).strip()
        if next_data_type.startswith("struct<") or next_data_type.startswith("array<"):
            d["array"] = _unpack_complex_etl_type(next_data_type)
        else:
            d["array"] = next_data_type
        return d
    else:
        return data_type


class EtlManagerConverter(BaseConverter):
    def __init__(self):
        """
        Converts metadata objects to etl-manager metadata
        and vice-versa.

        Note that this converter has no options
        (i.e. EtlManagerConverter().options returns
        the BaseCoverterOptions)

        Example:
        from mojap_metadata.converters.etl_manager_converter import (
            EtlManagerConverter,
        )
        emc = EtlManagerConverter()
        metadata = Metadata.from_json("my-table-metadata.json")
        table_meta = emc.generate_from_meta(metadata) # get TableMeta obj
        """

        super().__init__(None)
        self._default_type_converter = _default_type_converter
        self._reverse_default_type_converter = _reverse_default_type_converter

    def warn_conversion(self, coltype, converted_type, is_fully_supported):
        if converted_type is None:
            raise ValueError(
                f"{coltype} has no equivalent in Metadata so cannot be converted"
            )

        if not self.options.ignore_warnings and not is_fully_supported:
            w = (
                f"{coltype} is not fully supported by our Metadata using best "
                "representation. To supress these warnings set this converters "
                "options.ignore_warnings = True"
            )
            warnings.warn(w)

    def convert_col_type(self, coltype: str) -> str:
        """Converts our metadata types to Athena/Glue versions

        Args:
            coltype (str): str representation of our metadata column types

        Returns:
            str: String representation of athena column type version of `coltype`
        """

        data_type = _unpack_complex_data_type(coltype)

        return _flatten_and_convert_complex_data_type(
            data_type, self.convert_basic_col_type
        )

    def convert_basic_col_type(self, coltype: str):
        """Converts our metadata types (non complex ones)
        to etl-manager metadata. Used with the _flatten_and_convert_complex_data_type
        and convert_col_type functions.

        Args:
            coltype ([str]): str representation of our metadata column types

        Returns:
            str: String representation of etl-manager column type version of `coltype`
        """
        if coltype.startswith("decimal128"):
            t, is_supported = self._default_type_converter.get("decimal128")
            brackets = coltype.split("(")[1].split(")")[0]
            t = f"{t}({brackets})"
        elif coltype.startswith("binary"):
            coltype_ = coltype.split("(", 1)[0]
            t, is_supported = self._default_type_converter.get(coltype_, (None, None))
        else:
            t, is_supported = self._default_type_converter.get(coltype, (None, None))

        self.warn_conversion(coltype, t, is_supported)
        return t

    def convert_columns(
        self, metadata: Metadata, include_extra_column_params: bool = True
    ) -> List[dict]:
        """Converts metadata.columns to a list of etl_manager columns

        Args:
            metadata (Metadata): Metadata object
            include_extra_column_params (bool, optional): If true add on the extra
                column properties from metadata to the etl_manager meta data columns.

        Returns:
            List[dict]: list of etl_manager columns
        """

        etl_manager_cols = []
        etl_manager_params = ["nullable", "sensitive", "enum", "pattern"]

        default_params = ["name", "type", "description"]

        for c in metadata.columns:
            etl_manager_cols.append(
                {
                    "name": c["name"],
                    "type": self.convert_col_type(c["type"]),
                    "description": c.get("description", ""),
                }
            )

            for k, v in c.items():
                if k in default_params:
                    continue

                if include_extra_column_params or (k in etl_manager_params):
                    etl_manager_cols[-1][k] = v

        return etl_manager_cols

    def _default_file_format_to_data_format(self, file_format):

        if file_format.startswith("csv"):
            data_format = "csv"
        elif file_format.startswith("avro"):
            data_format = "avro"
        elif file_format.startswith("json") or file_format.startswith("ndjson"):
            data_format = "json"
        elif file_format.startswith("parquet"):
            data_format = "parquet"
        else:
            warnings.warn(
                "Could not determine data_format. Try setting your own "
                "file_format_mapper."
            )
            data_format = file_format
        return data_format

    def generate_from_meta(
        self,
        metadata: Metadata,
        table_location: str = None,
        file_format_mapper: Callable = None,
        include_extra_column_params: bool = True,
        glue_specific: dict = None,
    ) -> TableMeta:
        """Generates a TableMeta object from our metadata

        Args:
            metadata (Metadata): metadata object from the Metadata class
            file_format_mapper (Callable, optional): If not set the function
                will infer what the etl_manager data_type should be based on what
                your file format starts with. e.g. parquet.snappy -> parquet,
                csv.gz -> csv, etc. If you want to use your own mapper set a function
                object to this param e.g.
                file_format_mapper = my_lookup_dict.get
            table_location (str): relative path in S3, this is the location property of
                etl_manager metadata. If not set defaults to "<table_name>/"
            include_extra_column_params (bool, option): If True will add on additional
                column parameters that may not be used by etl_manager but other
                downstream tools that use etl_manager schemas. Set to False to only
                convert the params that are used by etl_manager.
            glue_specific (dict): dictionary used for specific glue parameterisation
                used by etl_manager
        Returns:
            TableMeta: An object from the TableMeta class in etl_manager.meta
        """

        if file_format_mapper:
            data_format = file_format_mapper(metadata.file_format)
        else:
            data_format = self._default_file_format_to_data_format(metadata.file_format)

        columns = self.convert_columns(metadata, include_extra_column_params)
        location = table_location if table_location else f"{metadata.name}/"

        tab = TableMeta(
            name=metadata.name,
            location=location,
            columns=columns,
            data_format=data_format,
            description=getattr(metadata, "description", ""),
            partitions=getattr(metadata, "partitions", []),
            primary_key=getattr(metadata, "primary_key", []),
            glue_specific=glue_specific,
            database=None,
        )

        return tab

    def reverse_convert_col_type(self, coltype: str):
        """Converts etl-manager metadata col types to Metadata col types

        Args:
            coltype ([str]): str representation of etl-manager column types

        Returns:
            [type]: str representation of Metadata col type
        """
        data_type = _unpack_complex_etl_type(coltype)

        return _flatten_and_convert_complex_data_type(
            data_type,
            self.reverse_convert_basic_col_type,
            complex_dtype_names=("array", "struct"),
        )

    def reverse_convert_basic_col_type(self, coltype: str):
        """Converts basic etl-manager metadata col types to Metadata col types

        Args:
            coltype ([str]): str representation of etl-manager column types

        Returns:
            [type]: str representation of Metadata col type
        """
        if coltype.startswith("decimal"):
            t, is_supported = self._reverse_default_type_converter.get("decimal")
            brackets = coltype.split("(")[1].split(")")[0]
            t = f"{t}({brackets})"
        elif coltype.startswith("binary"):
            coltype_ = coltype.split("(", 1)[0]
            t, is_supported = self._reverse_default_type_converter.get(
                coltype_, (None, None)
            )
        else:
            t, is_supported = self._reverse_default_type_converter.get(
                coltype, (None, None)
            )

        self.warn_conversion(coltype, t, is_supported)
        return t

    def generate_to_meta(
        self,
        table_meta: TableMeta,
        data_format_mapper: Callable = None,
        col_type_mapper: Callable = None,
    ) -> Metadata:
        """Takes a TableMeta object and converts it to our Metadata object

        Args:
            etl_manager_table_meta (Metadata): TableMeta object from etl-manager
            data_format_mapper (Callable, optional): If not set the function
                will just set the file_format parameter to the str in the
                original data_format of the TableMeta. If you want to use
                your own mapper set a function object to this param e.g.
                data_format_mapper = my_lookup_dict.get
            col_type_mapper (Callable, option): If not set the col type conversion
                from TableMeta -> Metadata is done using the converters
                reverse_convert_col_type method. If you need a custom conversion
                set a function to this parameter to use said function instead of
                reverse_convert_col_type This callable should expect the TableMeta
                col type str and return the Metadata col type str name.
        Returns:
            TableMeta: An object from the TableMeta class in etl_manager.meta
        """

        table_meta_dict = deepcopy(table_meta.to_dict())

        renamed_params = {"data_format": "file_format"}
        for old_name, new_name in renamed_params.items():
            table_meta_dict[new_name] = table_meta_dict.pop(old_name)

        if data_format_mapper:
            table_meta_dict["file_format"] = data_format_mapper(
                table_meta_dict["file_format"]
            )

        # remove etl_manager schema
        del table_meta_dict["$schema"]

        # convert columns
        etl_cols = table_meta_dict.pop("columns")
        for c in etl_cols:
            if col_type_mapper is None:
                c["type"] = self.reverse_convert_col_type(c["type"])
            else:
                c["type"] = col_type_mapper(c["type"])

        table_meta_dict["columns"] = etl_cols

        table_meta_dict["_converted_from"] = "etl_manager"
        return Metadata.from_dict(table_meta_dict)



class NonIcebergGlueTable(Exception):
    pass


class MalformedIcebergPartition(Exception):
    pass


class GlueIcebergTableExists(Exception):
    pass


class UnsupportedIcebergSchemaEvolution(Exception):
    pass



from typing import Union
from dataclasses import dataclass

from sqlalchemy import inspect
from sqlalchemy.engine import Engine, Connection
from sqlalchemy.types import (
    SmallInteger,
    Integer,
    BigInteger,
    REAL,
    Float,
    Numeric,
    String,
    DateTime,
    Date,
    Boolean,
    _Binary,
)

from mojap_metadata import Metadata
from mojap_metadata.converters import BaseConverter

_sqlalchemy_type_map = {
    SmallInteger: "int16",
    BigInteger: "int64",
    Integer: "int32",
    REAL: "float16",
    Float: "float64",
    String: "string",
    DateTime: "timestamp(ms)",
    Date: "date64",
    Boolean: "bool",
    _Binary: "binary",
}


@dataclass
class SQLAlchemyConverterOptions:
    default_decimal_precision: int = 38
    default_decimal_scale: int = 10


class SQLAlchemyConverter(BaseConverter):
    def __init__(
        self,
        connectable: Union[Engine, Connection],
        options: SQLAlchemyConverterOptions = None,
    ):
        """_Converts SQLAlchemy DDL to metadata object_

        Args:
            connectable (Union[Engine, Connection]): _A SQLAlchemy Engine or Connection_
        """
        if options is None:
            options = SQLAlchemyConverterOptions()

        super().__init__(options)
        self.connectable = connectable
        self.inspector = inspect(connectable)
        self._sqlalchemy_type_map = _sqlalchemy_type_map

    def _get_dtype(self, input_type) -> str:
        dtype = "string"
        for i, k in _sqlalchemy_type_map.items():
            if isinstance(input_type, i):
                dtype = k
                break
        return dtype

    def _convert_to_decimal(self, col_type) -> str:
        if col_type.precision is None:
            dp = self.options.default_decimal_precision
            ds = self.options.default_decimal_scale
            return f"decimal128({dp},{ds})"
        elif col_type.scale is None:
            return f"decimal128({str(col_type.precision)},0)"
        else:
            return f"decimal128({str(col_type.precision)},{str(col_type.scale)})"

    def _get_table_description(self, table, schema) -> str:
        description = ""
        try:
            description = (
                self.inspector.get_table_comment(table, schema=schema)["text"] or ""
            )
        except NotImplementedError:
            pass
        return description

    def convert_to_mojap_type(self, col_type) -> str:
        """_Converts a SQLAlchemy data type into a mojap data type_

        Args:
            col_type (_type_): _A SQLAlchemy data type_

        Returns:
            str: _A mojap data type_
        """
        if isinstance(col_type, Numeric) and not isinstance(col_type, Float):
            dtype = self._convert_to_decimal(col_type)
        else:
            dtype = self._get_dtype(col_type)
        return dtype

    def generate_to_meta(self, table: str, schema: str) -> Metadata:
        """_Generates a Metadata object from a specified table and schema_

        Args:
            table (str): _Table name to generate the metadata for_
            schema (str): _Schema name to generate the metadata for_

        Returns:
            Metadata: _Metadata object_
        """
        rows = self.inspector.get_columns(table, schema)
        columns = []
        for col in rows:
            columns.append(
                {
                    "name": col["name"].lower(),
                    "type": self.convert_to_mojap_type(col["type"]),
                    "description": col.get("comment") or "",
                    "nullable": col.get("nullable"),
                }
            )
        pk = self.inspector.get_pk_constraint(table, schema)
        d = {
            "name": table,
            "columns": columns,
            "primary_key": pk["constrained_columns"],
            "database": schema,
            "description": self._get_table_description(table, schema),
        }

        meta_output = Metadata.from_dict(d)
        return meta_output

    def generate_to_meta_list(self, schema: str) -> list:
        """_Generates a list of Metadata objects for all the tables in a schema_

        Args:
            schema (str): _Schema name to generate the metadata for_

        Returns:
            list: _list of Metadata objects_
        """
        schema_metadata = []
        table_names = self.inspector.get_table_names(schema)
        table_names = sorted(table_names)
        for table in table_names:
            table_metadata = self.generate_to_meta(table, schema)
            schema_metadata.append(table_metadata)
        return schema_metadata



import boto3
import json
import os
import re
import time
import warnings

import importlib.resources as pkg_resources

from dataclasses import dataclass
from mojap_metadata.converters import (
    BaseConverter,
    _flatten_and_convert_complex_data_type,
)
from mojap_metadata.metadata.metadata import (
    Metadata,
    _unpack_complex_data_type,
    _metadata_complex_dtype_names,
)
from mojap_metadata.converters.glue_converter import specs
from typing import Tuple, List, Union


# Format generictype: (glue_type, is_fully_supported)
# If not fully supported we decide on best option
# if glue_type is Null then we have no way to safely
# convert it
_default_type_converter = {
    "null": (None, False),
    "bool": ("boolean", True),
    "bool_": ("boolean", True),
    "int8": ("tinyint", True),
    "int16": ("smallint", True),
    "int32": ("int", True),
    "int64": ("bigint", True),
    "uint8": ("smallint", False),
    "uint16": ("int", False),
    "uint32": ("bigint", False),
    "uint64": (None, False),
    "decimal128": ("decimal", True),
    "float16": ("float", False),
    "float32": ("float", True),
    "float64": ("double", True),
    "time32": (None, False),
    "time32(s)": (None, False),
    "time32(ms)": (None, False),
    "time64(us)": (None, False),
    "time64(ns)": (None, False),
    "date32": ("date", True),
    "date64": ("date", True),
    "timestamp(s)": ("timestamp", True),
    "timestamp(ms)": ("timestamp", True),
    "timestamp(us)": ("timestamp", True),
    "timestamp(ns)": ("timestamp", True),
    "string": ("string", True),
    "large_string": ("string", True),
    "utf8": ("string", True),
    "large_utf8": ("string", True),
    "binary": ("binary", True),
    "large_binary": ("binary", True),
    "list": ("array", True),
    "list_": ("array", True),
    "large_list": ("array", True),
    "struct": ("struct", True),
}

_glue_to_mojap_type_converter = {
    "boolean": ("bool", True),
    "tinyint": ("int8", True),
    "smallint": ("int16", False),
    "int": ("int32", False),
    "integer": ("int32", False),
    "bigint": ("int64", False),
    "double": ("float64", True),
    "float": ("float32", False),
    "decimal": ("decima128", True),
    "char": ("string", True),
    "varchar": ("string", True),
    "string": ("string", True),
    "binary": ("binary", False),
    "date": ("date64", False),
    "timestamp": ("timestamp(s)", False),
    "array": ("large_list", False),
    "struct": ("struct", False),
}


@dataclass
class CsvOptions:
    """
    Specific options for CSV spec
    """

    serde: str = "lazy"
    skip_header: bool = False
    sep: str = ","
    quote_char: str = '"'
    escape_char: str = "\\"
    compressed: bool = False


@dataclass
class JsonOptions:
    serde: str = "hive"
    compressed: bool = False


@dataclass
class ParquetOptions:
    # compression:str = "SNAPPY"
    compressed: bool = True


SpecOptions = Union[CsvOptions, JsonOptions, ParquetOptions]


@dataclass
class GlueConverterOptions:
    """
    Options Class for the GlueConverter

    deafultcsv_serde (str):
      Jinja template that is used to generate CSV ddl. Defaults to
      lazy serde template. To use the open serde template you can use
      options.set_csv_serde("open").

    json_serde (str):
      Jinja template that is used to generate JSON ddl. Can be set to
      a custom json ddl template but defaults to our recommended one:
      `specs/json_ddl.txt`.

    parquet_template (str):
      Jinja template that is used to generate PARQUET ddl. Can be set to
      a custom json ddl template but defaults to our recommended one:
      `specs/parquet_ddl.txt`.

    default_db_name (str):
      Default database name to default to when defining which database
      the table belongs to. Used when calling `GlueConverter.generate_from_meta`
      method and no database_name is specified.

    default_db_base_path (str):
      Default s3 base path default to when defining the the table exists in S3.
      Used when calling `GlueConverter.generate_from_meta` method and no
      table_location is  specified. When no table_location is specified, the
      output DDL wil define the table location as <default_db_base_path>/<table_name>/.

    ignore_warnings (bool, default=False):
      If converter should not warning users of imperfect type conversions.

    skip_header (bool):
      parameter to csv ddl function

    sep (str):
      parameter to csv ddl function

    quote_char (str):
      parameter to csv ddl function

    escape_char (str):
      parameter to csv ddl function

    line_term_char (str):
      parameter to csv ddl function

    parquet_compression (str):
      parameter to parquet ddl function
    """

    csv = CsvOptions()
    json = JsonOptions()
    parquet = ParquetOptions()
    default_db_name: str = None
    default_db_base_path: str = None
    ignore_warnings: bool = False

    def set_csv_serde(self, serde_name: str):
        allowed_serdes = ["lazy", "open"]
        if serde_name not in allowed_serdes:
            err_msg = (
                f"Input serde_name must be one of {allowed_serdes} "
                f"but got {serde_name}."
            )
            raise ValueError(err_msg)
        else:
            self.csv.serde = serde_name

    def set_json_serde(self, serde_name: str):
        allowed_serdes = ["hive", "openx"]
        if serde_name not in allowed_serdes:
            err_msg = (
                f"Input serde_name must be one of {allowed_serdes} "
                f"but got {serde_name}."
            )
            raise ValueError(err_msg)
        else:
            self.json.serde = serde_name


class GlueConverter(BaseConverter):
    def __init__(self, options: GlueConverterOptions = None):
        """
        Converts metadata objects to a Hive DDL.

        options (GlueConverterOptions, optional): See ?GlueConverterOptions
        for more details. If not set a default GlueConverterOptions is set to
        the options parameter.

        Example:
        from mojap_metadata.converters.glue_converter import (
            GlueConverter,
            GlueConverterOptions,
            _create_open_csv_ddl,
        )
        options = GlueConverterOptions(csv_ddl = _create_open_csv_ddl)
        gc = GlueConverter(options)
        metadata = Metadata.from_json("my-table-metadata.json")
        ddl = gc.generate_from_meta(metadata) # get Glue/Hive DDL
        """
        if options is None:
            options = GlueConverterOptions()

        super().__init__(options)
        self._default_type_converter = _default_type_converter

    def warn_conversion(self, coltype, converted_type, is_fully_supported):
        if converted_type is None:
            raise ValueError(
                f"{coltype} has no equivalent in Athena/Glue " "so cannot be converted"
            )

        if not self.options.ignore_warnings and not is_fully_supported:
            w = (
                f"{coltype} is not fully supported by Athena using best "
                "representation. To supress these warnings set this converters "
                "options.ignore_warnings = True"
            )
            warnings.warn(w)

    def convert_col_type(self, coltype: str) -> str:
        """Converts our metadata types to Athena/Glue versions

        Args:
            coltype (str): str representation of our metadata column types

        Returns:
            str: String representation of athena column type version of `coltype`
        """

        data_type = _unpack_complex_data_type(coltype)

        return _flatten_and_convert_complex_data_type(
            data_type, self.convert_basic_col_type, field_sep=","
        )

    def convert_basic_col_type(self, coltype: str) -> str:
        """Converts our metadata types (non complex ones)
        to etl-manager metadata. Used with the _flatten_and_convert_complex_data_type
        and convert_col_type functions.

        Args:
            coltype (str): str representation of our metadata column types

        Returns:
            str: String representation of athena column type version of `coltype`
        """
        if coltype.startswith("decimal128"):
            t, is_supported = self._default_type_converter.get("decimal128")
            brackets = coltype.split("(")[1].split(")")[0]
            t = f"{t}({brackets})"
        elif coltype.startswith("binary"):
            coltype_ = coltype.split("(", 1)[0]
            t, is_supported = self._default_type_converter.get(coltype_, (None, None))
        else:
            t, is_supported = self._default_type_converter.get(coltype, (None, None))

        self.warn_conversion(coltype, t, is_supported)
        return t

    def convert_columns(self, metadata: Metadata) -> Tuple[List, List]:
        cols = []
        partitions = []
        for c in metadata.columns:
            if c["name"] in metadata.partitions:
                partitions.append(
                    {"Name": c["name"], "Type": self.convert_col_type(c["type"])}
                )
                if "description" in c:
                    partitions[-1]["Comment"] = c["description"]
            else:
                cols.append(
                    {"Name": c["name"], "Type": self.convert_col_type(c["type"])}
                )
                if "description" in c:
                    cols[-1]["Comment"] = c["description"]
        return cols, partitions

    def generate_from_meta(
        self,
        metadata: Metadata,
        database_name: str = None,
        table_location: str = None,
    ) -> dict:
        """Generates the Hive DDL from our metadata

        Args:
            metadata (Metadata): metadata object from the Metadata class
            database_name (str, optional): database name needed for table DDL.
              If `None` this function will look to the options.default_database_name
              attribute to find a name. Defaults to None.
            table_location (str, optional): S3 location of where table is stored
              needed for table DDL. If `None` this function will look to the
              options.default_database_name attribute to find a name. Defaults
              to None.
        Raises:
            ValueError: If database_name and table_location are not set (and there
              are no default options set)

        Returns:
            dict: for Glue API
        """

        # set database_name to metadata.database_name if none
        database_name = database_name if database_name else metadata.database_name
        # do the same with table_location
        table_location = table_location if table_location else metadata.table_location

        ff = metadata.file_format
        if ff.startswith("csv"):
            opts = self.options.csv

        elif ff.startswith("json"):
            opts = self.options.json

        elif ff.startswith("parquet"):
            opts = self.options.parquet

        else:
            raise ValueError(
                f"No ddl template for type: {ff} in options "
                "(only supports formats starting with csv, json or parquet)"
            )

        if not database_name:
            if self.options.default_db_name:
                database_name = self.options.default_db_name
            else:
                error_msg = (
                    "Either set database_name in the function "
                    "or set a default_db_name in the GlueConverter.options"
                )
                raise ValueError(error_msg)

        if not table_location:
            if self.options.default_db_base_path:
                table_location = os.path.join(
                    self.options.default_db_base_path, f"{metadata.name}"
                )
            else:
                error_msg = (
                    "Either set table_location in the function "
                    "or set a database_base_path in the "
                    "GlueConverter.options.database_base_path"
                )
                raise ValueError(error_msg)

        table_cols, partition_cols = self.convert_columns(metadata)

        spec = generate_spec_from_template(
            database_name=database_name,
            table_name=metadata.name,
            location=table_location,
            spec_opts=opts,
            table_desc=metadata.description,
            columns=table_cols,
            partitions=partition_cols,
        )
        return spec


class GlueTable(BaseConverter):
    def __init__(self, glue_converter_options: GlueConverterOptions = None):
        super().__init__(None)
        self.gc = GlueConverter(glue_converter_options)

    def convert_basic_col_type(self, col_type: str):
        if col_type.startswith("decimal"):
            regex = re.compile(r"decimal(\(\d+,( |)\d+\)|\(\d+\))")
            bracket_numbers = regex.match(col_type).groups()[0]
            return f"decimal128{bracket_numbers}"
        elif col_type.startswith(("char", "varchar")):
            coltype_ = col_type.split("(", 1)[0]
            return _glue_to_mojap_type_converter[coltype_][0]
        else:
            return _glue_to_mojap_type_converter[col_type][0]

    def convert_col_type(self, col_type: str):
        data_type = _unpack_complex_data_type(col_type)
        return _flatten_and_convert_complex_data_type(
            data_type, self.convert_basic_col_type, field_sep=","
        )

    def convert_columns(self, columns: List[dict]) -> List[dict]:
        mojap_meta_cols = []
        for col in columns:
            mojap_col_type = self.convert_col_type(col["Type"])
            if mojap_col_type.startswith(_metadata_complex_dtype_names):
                normalised_col_type = col["Type"].split("<", 1)[0]
            else:
                normalised_col_type = col["Type"].split("(", 1)[0]
            full_support = _glue_to_mojap_type_converter.get(normalised_col_type)[1]
            if not full_support:
                warnings.warn(
                    f"type {col['Type']} not fully supported, "
                    "likely due to multiple conversion options"
                )
            # create the column in mojap meta format
            meta_col = {
                "name": col["Name"],
                "type": mojap_col_type,
            }
            if col.get("Comment"):
                meta_col["description"] = col.get("Comment")
            mojap_meta_cols.append(meta_col)
        return mojap_meta_cols

    def generate_from_meta(
        self,
        metadata: Union[Metadata, str, dict],
        database_name: str = None,
        table_location: str = None,
        run_msck_repair: bool = False,
    ):
        """
        Creates a glue table from metadata
        arguments:
            - metadata: Metadata object, string path, or dictionary metadata.
            - database_name (optional): name of the glue database the table is to be
            created in. can also be a property of the metadata.
            - table_location (optional): the s3 location of the table. can also be a
            property of the metadata.
            - run_msck_repair (optional): run msck repair table on the created table,
            should be set to True for tables with partitions.
        Raises:
            - ValueError if run_msck_repair table is False, metadata has partitions, and
            options.ignore_warnings is set to False
        """

        # set database_name to metadata.database_name if none
        database_name = database_name if database_name else metadata.database_name
        # do the same with table_location
        table_location = table_location if table_location else metadata.table_location

        glue_client = boto3.client(
            "glue",
            region_name=os.getenv(
                "AWS_REGION", os.getenv("AWS_DEFAULT_REGION", "eu-west-1")
            ),
        )
        metadata = Metadata.from_infer(metadata)
        boto_dict = self.gc.generate_from_meta(
            metadata, database_name=database_name, table_location=table_location
        )
        # create database if it doesn't exist
        _start_query_execution_and_wait(
            database_name, f"CREATE DATABASE IF NOT EXISTS {database_name};"
        )

        # delete table if it exists
        try:
            glue_client.delete_table(DatabaseName=database_name, Name=metadata.name)
        except glue_client.exceptions.EntityNotFoundException:
            pass

        glue_client.create_table(**boto_dict)

        if (
            not run_msck_repair
            and metadata.partitions
            and not self.options.ignore_warnings
        ):
            w = (
                "metadata has partitions and run_msck_repair is set to false. To "
                "To supress these warnings set this converters "
                "options.ignore_warnings = True"
            )
            warnings.warn(w)
        elif run_msck_repair:
            _start_query_execution_and_wait(
                database_name, f"msck repair table {database_name}.{metadata.name}"
            )

    def generate_to_meta(self, database: str, table: str) -> Metadata:
        # get the table information
        glue_client = boto3.client("glue")
        resp = glue_client.get_table(DatabaseName=database, Name=table)

        # pull out just the columns
        columns = resp["Table"]["StorageDescriptor"]["Columns"]

        # convert the columns
        mojap_meta_cols = self.convert_columns(columns)

        # check for partitions
        partitions = resp["Table"].get("PartitionKeys")
        if partitions:
            # convert
            part_cols_full = self.convert_columns(partitions)
            # extend the mojap_meta_cols with the partiton cols
            mojap_meta_cols.extend(part_cols_full)
            part_cols = [p["name"] for p in part_cols_full]

            # make a metadata object
            meta = Metadata(name=table, columns=mojap_meta_cols, partitions=part_cols)
        else:
            meta = Metadata(name=table, columns=mojap_meta_cols)

        # get the file format if possible
        try:
            ff = resp["Table"]["StorageDescriptor"]["Parameters"].get("classification")
        except KeyError:
            warnings.warn("unable to parse file format, please manually set")
            ff = None

        if ff:
            meta.file_format = ff.lower()

        return meta


def _get_base_table_spec(spec_name: str, serde_name: str = None) -> dict:
    """Gets a table spec (dict) for a specific name
    prefilled with standard properties / info for that
    specific spec.

    Args:
        spec_name (str): Name of the spec currently -
        'csv', 'json' or 'parquet'

        serde_name (str): Name of the specific serde -
        CSV: 'open' or 'lazy'
        JSON: 'hive' or 'openx'
        PARQUET: None

    Returns:
        dict: A base spec that can be used with boto to create a table.
        Once specific details from metadata are filled into it.
    """
    if serde_name:
        filename = f"{serde_name}_{spec_name}_spec.json"
    else:
        filename = f"{spec_name}_spec.json"

    table_spec = json.load(pkg_resources.open_text(specs, filename))
    return table_spec


def _get_spec_and_serde_name_from_opts(spec_opts) -> Tuple[str, str]:
    """Returns the spec name and serde name for a given option Class
    and parameters

    Args:
        spec_opts ([type]): [description]

    Raises:
        ValueError: [description]

    Returns:
        Tuple[str, str]: [description]
    """
    if isinstance(spec_opts, CsvOptions):
        spec_name = "csv"
        serde_name = spec_opts.serde
    elif isinstance(spec_opts, JsonOptions):
        spec_name = "json"
        serde_name = spec_opts.serde
    elif isinstance(spec_opts, ParquetOptions):
        spec_name = "parquet"
        serde_name = None
    else:
        raise ValueError(
            f"expected opts to be of an options Type not {type(spec_opts)}"
        )

    return spec_name, serde_name


def _convert_opts_into_dict(spec_opts: SpecOptions):
    """Takes the spec_opts and converts it to a dict
    Just used to pass to Template.render().

    Args:
        spec_opts (SpecOptions): One of the SpecOptions
        classes

    Returns:
        dict: A dict representation of the data class
    """
    out_dict = {}
    for k, _ in spec_opts.__annotations__:
        out_dict[k] = getattr(spec_opts, k)
    return out_dict


def _start_query_execution_and_wait(db: str, sql: str):
    ath = boto3.client("athena")
    QueryExecutionContext = {"Database": db}
    WorkGroup = "primary"
    res = ath.start_query_execution(
        QueryString=sql,
        QueryExecutionContext=QueryExecutionContext,
        WorkGroup=WorkGroup,
    )
    query_exec_id = res["QueryExecutionId"]
    while response := ath.get_query_execution(QueryExecutionId=query_exec_id):
        state = response["QueryExecution"]["Status"]["State"]
        if state not in ["SUCCEEDED", "FAILED"]:
            time.sleep(0.25)
        else:
            break

    if not state == "SUCCEEDED":
        raise ValueError(response["QueryExecution"]["Status"].get("StateChangeReason"))


def generate_spec_from_template(
    database_name,
    table_name,
    location,
    spec_opts: SpecOptions,
    table_desc="",
    columns=[],
    partitions=[],
):
    spec_name, serde_name = _get_spec_and_serde_name_from_opts(spec_opts)

    base_spec = _get_base_table_spec(spec_name, serde_name)
    base_spec["Name"] = table_name
    base_spec["Description"] = table_desc
    base_spec["StorageDescriptor"]["Columns"] = columns
    base_spec["PartitionKeys"] = partitions
    base_spec["StorageDescriptor"]["Location"] = location

    # Do general options
    base_spec["StorageDescriptor"]["Compressed"] = spec_opts.compressed

    # Do CSV options
    if spec_name == "csv":
        csv_param_lu = {
            "sep": {"lazy": "field.delim", "open": "separatorChar"},
            "quote_char": {"lazy": None, "open": "quoteChar"},
            "escape_char": {"lazy": "escape.delim", "open": "escapeChar"},
        }

        if spec_opts.skip_header:
            (
                base_spec["StorageDescriptor"]["SerdeInfo"]["Parameters"][
                    "skip.header.line.count"
                ]
            ) = "1"

        if spec_opts.sep:
            param_name = csv_param_lu["sep"][serde_name]
            (
                base_spec["StorageDescriptor"]["SerdeInfo"]["Parameters"][param_name]
            ) = spec_opts.sep

        if spec_opts.quote_char and serde_name != "lazy":
            (
                base_spec["StorageDescriptor"]["SerdeInfo"]["Parameters"]["quoteChar"]
            ) = spec_opts.quote_char

        if spec_opts.escape_char:
            param_name = csv_param_lu["escape_char"][serde_name]
            (
                base_spec["StorageDescriptor"]["SerdeInfo"]["Parameters"][param_name]
            ) = spec_opts.escape_char

    # Do JSON options
    if spec_name == "json":
        json_col_paths = ",".join([c["Name"] for c in columns])
        (
            base_spec["StorageDescriptor"]["SerdeInfo"]["Parameters"]["paths"]
        ) = json_col_paths

    out_dict = {"DatabaseName": database_name, "TableInput": base_spec}
    return out_dict



import sqlalchemy


def list_schemas(connection: sqlalchemy.engine.Engine):
    """List non-system schemas in a database."""
    response = connection.execute(
        """
        SELECT schema_name
        FROM information_schema.schemata
        """
    ).fetchall()
    system_schemas = (
        "pg_catalog",
        "information_schema",
        "pg_toast",
        "pg_temp_1",
        "pg_toast_temp_1",
    )
    return [r[0] for r in response if r[0] not in system_schemas]


def list_tables(connection: sqlalchemy.engine.Engine, schema: str = "public"):
    """List tables in a database."""
    # WHERE schemaname != 'pg_catalog' AND schemaname != 'information_schema'
    response = connection.execute(
        f"""
        SELECT tablename
        FROM pg_catalog.pg_tables
        WHERE schemaname = '{schema}'
        """
    ).fetchall()
    return [r[0] for r in response]


def list_dbs(connection: sqlalchemy.engine.Engine):
    """List databases from a connectionection."""
    response = connection.execute(
        """
        SELECT datname
        FROM pg_database
        """
    ).fetchall()
    return [r[0] for r in response]


def list_meta_data(
    connection: sqlalchemy.engine.Engine, table_name: str, schema: str
) -> list:
    """List metadata  for  table in a particular schema"""

    response = connection.execute(
        f""" SELECT
            cols.column_name, cols.data_type,cols.is_nullable,
            col_description((table_schema||'.'||table_name)::regclass::oid,
            ordinal_position) as column_comment
            FROM information_schema.columns cols
            WHERE
                cols.table_schema  = '{schema}' AND
                cols.table_name    = '{table_name}';"""
    )
    rows = response.fetchall()
    cols = response.keys()
    return rows, list(cols)



from mojap_metadata.metadata.metadata import (
    Metadata,
    _metadata_complex_dtype_names,
)
from typing import Union, Any, Callable, Tuple
from collections.abc import Mapping
from dataclasses import dataclass


# https://gist.github.com/angstwad/bf22d1822c38a92ec0a9
def _dict_merge(dct, merge_dct):
    """Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.
    :param dct: dict onto which the merge is executed
    :param merge_dct: dct merged into dct
    :return: None
    """
    for k, v in merge_dct.items():
        if k in dct and isinstance(dct[k], dict) and isinstance(merge_dct[k], Mapping):
            _dict_merge(dct[k], merge_dct[k])
        else:
            dct[k] = merge_dct[k]


def _flatten_and_convert_complex_data_type(
    data_type: Union[dict, str],
    converter_fun: Callable,
    complex_dtype_names: Tuple[str] = None,
    field_sep: str = ", ",
) -> str:
    """Recursive function to flattern a complex datatype in a dictionary
    format i.e. output from (from Metadata.unpack_complex_data_type).
    And flattern it down to a string but with the converted data types.

    Args:
        data_type (dict): complex data type as a dictionary
        converter_fun (Callable): standard converter function to change
            a str data_type to the new data_type
        complex_dtype_names (Tuple[str]): A set of names that define a complex
            datatype (the name given before <>). If None, defaults to
            _metadata_complex_dtype_names from the metadata module
            which are the agnostic dtype names

    Returns:
        str: Complex datatype converted back into a flattened str of
            converted datatypes
    """
    if complex_dtype_names is None:
        complex_dtype_names = _metadata_complex_dtype_names

    if isinstance(data_type, str):
        return converter_fun(data_type)

    else:
        fields = []
        for k, v in data_type.items():
            if k in complex_dtype_names:
                inner_data_type = _flatten_and_convert_complex_data_type(
                    v, converter_fun, complex_dtype_names, field_sep=field_sep
                )
                return f"{converter_fun(k)}<{inner_data_type}>"
            else:
                new_v = _flatten_and_convert_complex_data_type(
                    v, converter_fun, complex_dtype_names, field_sep=field_sep
                )
                fields.append(f"{k}:{new_v}")
                del new_v
        return field_sep.join(fields)


@dataclass
class BaseConverterOptions:
    ignore_warnings = False


class BaseConverter:
    def __init__(self, options: Union[BaseConverterOptions, Any] = None):
        """
        Base class to be used as standard for parsing in an object, say DDL
        or oracle db connection and then outputting a Metadata class. Not sure
        if needed or will be too strict for generalisation.

        options (BaseConverterOptions): A simple class that lets users set or get
        particular paramters. Each one will specific to the converter but each
        converter uses this standard class to access and set parameters.
        """
        if options is None:
            self.options = BaseConverterOptions()
        else:
            self.options = options

    def generate_to_meta(self, item, **kwargs) -> Metadata:
        """
        Should be overwritten to transform item into the Metadata object
        """
        raise NotImplementedError("This function has not been overwritten")

    def generate_from_meta(self, metadata: Metadata, **kwargs) -> object:
        """
        Should be overwritten to transform metadata into the MetaData object
        """
        raise NotImplementedError("This function has not been overwritten")

    def convert_col_type(self, coltype: str) -> Any:
        """
        Should be overwritten to transform the col type (str) from the our agnostic
        metadata types to the equivalent type for the converter
        """
        raise NotImplementedError("This function has not been overwritten")

    def reverse_convert_col_type(self, coltype: Any) -> str:
        """
        Should be overwritten to transform a coltype object (Any) to our agnostic
        metadata types
        """
        raise NotImplementedError("This function has not been overwritten")



import re

from mojap_metadata.metadata.metadata import Metadata, MetadataProperty


class IcebergMetadata(Metadata):
    table_type = MetadataProperty()
    noncurrent_columns = MetadataProperty()
    partition_transforms = MetadataProperty()

    def validate(self):
        super().validate()
        self._validate_partition_transform_is_supported()
        self._validate_same_number_of_partitions_as_transforms_if_given()

    def _validate_same_number_of_partitions_as_transforms_if_given(self):
        if self._data.get("partition_transforms") is not None and not self._data.get(
            "partition_transforms"
        ):
            if len(self._data["partitions"]) != len(self._data["partition_transforms"]):
                raise ValueError(
                    "The number of partitions and partition transforms do not match"
                )

    def _validate_partition_transform_is_supported(
        self,
        transform_regex: str = "|".join(
            [
                "(identity)",
                "(year)",
                "(day)",
                "(hour)",
                "(bucket\\[[0-9]*\\])",
                "(truncate\\[[0-9]*\\])",
            ]
        ),
    ):
        for transform in self._data.get("partition_transforms", []):
            matches = re.search(
                transform_regex,
                transform,
            )
            if matches is None:
                raise ValueError(
                    f"{transform} does not match pattern {transform_regex}"
                )



from typing import DefaultDict
import sqlalchemy

from mojap_metadata import Metadata
import mojap_metadata.converters.postgres_converter.postgres_functions as pg
from mojap_metadata.converters import BaseConverter

_default_type_converter = {
    "int8": "int8",
    "int16": "int16",
    "int32": "int32",
    "int64": "int64",
    "bigint": "int64",
    "int2": "int32",
    "int4": "int32",
    "integer": "int32",
    "smallint": "int32",
    "numeric": "float64",
    "double precision": "float64",
    "text": "string",
    "uuid": "string",
    "character": "string",
    "tsvector": "string",
    "jsonb": "string",
    "varchar": "string",
    "bpchar": "string",
    "date": "date64",
    "boolean": "bool",
    "timestamptz": "timestamp(ms)",
    "timestamp": "timestamp(ms)",
    "datetime": "timestamp(ms)",
    "bool": "bool",
}


class PostgresConverter(BaseConverter):
    def __init__(self):
        """
        Extracts and converts metadata to Metadata format
        """

        super().__init__()
        self._default_type_converter = _default_type_converter

    def convert_to_mojap_type(self, col_type: str) -> str:
        """Converts our postgress datatypes to mojap-metadata types

        Args:
            ct (str): str representation of postgres column types

        Returns:
            str: String representation of our metadata column types
        """

        output = (
            "string"
            if self._default_type_converter.get(col_type) is None
            else self._default_type_converter.get(col_type)
        )
        return output

    def get_object_meta(
        self, connection: sqlalchemy.engine.Engine, table: str, schema: str
    ) -> Metadata:
        """Extracts metadata from table and converts to mojap metadata format

        Args:
            connection: Database connection
            table: table name
            schema: schema name

        Returns:
            Metadata: Metadata object
        """

        rows = pg.list_meta_data(connection, table, schema)
        columns = []

        for col in rows[0]:

            column_type = self.convert_to_mojap_type(str(col[1]))
            columns.append(
                {
                    "name": col[0].lower(),
                    "type": column_type,
                    "description": None if str(col[3]) is None else str(col[3]),
                    "nullable": True if col[2] == "YES" else False,
                }
            )

        d = {"name": table, "columns": columns}

        meta_output = Metadata.from_dict(d)
        return meta_output

    def generate_from_meta(self, connection: sqlalchemy.engine.Engine) -> dict():
        """Extracts metadata for all the schema and tables and returns a list
        of Metadata

        Args:
            connection: Database connection with database details

        Returns:
            Metadata: Metadata object
        """
        meta_list_per_schema = DefaultDict(list)

        schema_names = pg.list_schemas(
            connection
        )  # database name will be passed on in the connection

        for schema in sorted(schema_names):
            table_names = pg.list_tables(connection, schema)

            for table in table_names:
                meta_output = self.get_object_meta(connection, table, schema)
                meta_list_per_schema[f"schema: {schema}"].append(meta_output)

        return meta_list_per_schema
